@@ -3,12 +3,12 @@
  * parse_relation.c
  *	  parser support routines dealing with relations
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/parser/parse_relation.c,v 1.90.2.2 2005/05/29 17:10:52 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/parser/parse_relation.c,v 1.102.4.1 2005/05/29 17:10:35 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -41,11 +41,11 @@ static Node *scanNameSpaceForRelid(ParseState *pstate, Node *nsnode,
 					  Oid relid);
 static void scanNameSpaceForConflict(ParseState *pstate, Node *nsnode,
 						 RangeTblEntry *rte1, const char *aliasname1);
-static Node *scanRTEForColumn(ParseState *pstate, RangeTblEntry *rte,
-				 char *colname);
 static bool isForUpdate(ParseState *pstate, char *refname);
-static bool get_rte_attribute_is_dropped(RangeTblEntry *rte,
-							 AttrNumber attnum);
+static void expandRelation(Oid relid, Alias *eref,
+			   int rtindex, int sublevels_up,
+			   bool include_dropped,
+			   List **colnames, List **colvars);
 static int	specialAttNum(const char *attname);
 static void warnAutoRange(ParseState *pstate, RangeVar *relation);
 
@@ -176,7 +176,7 @@ scanNameSpaceForRefname(ParseState *pstate, Node *nsnode,
 	}
 	else if (IsA(nsnode, List))
 	{
-		List	   *l;
+		ListCell   *l;
 
 		foreach(l, (List *) nsnode)
 		{
@@ -251,7 +251,7 @@ scanNameSpaceForRelid(ParseState *pstate, Node *nsnode, Oid relid)
 	}
 	else if (IsA(nsnode, List))
 	{
-		List	   *l;
+		ListCell   *l;
 
 		foreach(l, (List *) nsnode)
 		{
@@ -323,7 +323,7 @@ checkNameSpaceConflicts(ParseState *pstate, Node *namespace1,
 	}
 	else if (IsA(namespace1, List))
 	{
-		List	   *l;
+		ListCell   *l;
 
 		foreach(l, (List *) namespace1)
 			checkNameSpaceConflicts(pstate, lfirst(l), namespace2);
@@ -380,7 +380,7 @@ scanNameSpaceForConflict(ParseState *pstate, Node *nsnode,
 	}
 	else if (IsA(nsnode, List))
 	{
-		List	   *l;
+		ListCell   *l;
 
 		foreach(l, (List *) nsnode)
 			scanNameSpaceForConflict(pstate, lfirst(l), rte1, aliasname1);
@@ -399,7 +399,7 @@ int
 RTERangeTablePosn(ParseState *pstate, RangeTblEntry *rte, int *sublevels_up)
 {
 	int			index;
-	List	   *temp;
+	ListCell   *l;
 
 	if (sublevels_up)
 		*sublevels_up = 0;
@@ -407,9 +407,9 @@ RTERangeTablePosn(ParseState *pstate, RangeTblEntry *rte, int *sublevels_up)
 	while (pstate != NULL)
 	{
 		index = 1;
-		foreach(temp, pstate->p_rtable)
+		foreach(l, pstate->p_rtable)
 		{
-			if (rte == (RangeTblEntry *) lfirst(temp))
+			if (rte == (RangeTblEntry *) lfirst(l))
 				return index;
 			index++;
 		}
@@ -424,6 +424,45 @@ RTERangeTablePosn(ParseState *pstate, RangeTblEntry *rte, int *sublevels_up)
 	return 0;					/* keep compiler quiet */
 }
 
+/*
+ * Given an RT index and nesting depth, find the corresponding RTE.
+ * This is the inverse of RTERangeTablePosn.
+ */
+RangeTblEntry *
+GetRTEByRangeTablePosn(ParseState *pstate,
+					   int varno,
+					   int sublevels_up)
+{
+	while (sublevels_up-- > 0)
+	{
+		pstate = pstate->parentParseState;
+		Assert(pstate != NULL);
+	}
+	Assert(varno > 0 && varno <= list_length(pstate->p_rtable));
+	return rt_fetch(varno, pstate->p_rtable);
+}
+
+/*
+ * GetLevelNRangeTable
+ *	  Get the rangetable list for the N'th query level up from current.
+ */
+List *
+GetLevelNRangeTable(ParseState *pstate, int sublevels_up)
+{
+	int			index = 0;
+
+	while (pstate != NULL)
+	{
+		if (index == sublevels_up)
+			return pstate->p_rtable;
+		index++;
+		pstate = pstate->parentParseState;
+	}
+
+	elog(ERROR, "rangetable not found (internal error)");
+	return NIL;					/* keep compiler quiet */
+}
+
 /*
  * scanRTEForColumn
  *	  Search the column names of a single RTE for the given name.
@@ -437,28 +476,25 @@ RTERangeTablePosn(ParseState *pstate, RangeTblEntry *rte, int *sublevels_up)
  * nothing.  It might seem that we need to propagate the mark to all the
  * contained RTEs, but that is not necessary.  This is so because a join
  * expression can only appear in a FROM clause, and any table named in
- * FROM will be marked checkForRead from the beginning.
+ * FROM will be marked as requiring read access from the beginning.
  */
-static Node *
+Node *
 scanRTEForColumn(ParseState *pstate, RangeTblEntry *rte, char *colname)
 {
 	Node	   *result = NULL;
 	int			attnum = 0;
-	List	   *c;
+	ListCell   *c;
 
 	/*
 	 * Scan the user column names (or aliases) for a match. Complain if
 	 * multiple matches.
 	 *
-	 * Note: because eref->colnames may include names of dropped columns, we
-	 * need to check for non-droppedness before accepting a match. This
-	 * takes an extra cache lookup, but we can skip the lookup most of the
-	 * time by exploiting the knowledge that dropped columns are assigned
-	 * dummy names starting with '.', which is an unusual choice for
-	 * actual column names.
+	 * Note: eref->colnames may include entries for dropped columns, but
+	 * those will be empty strings that cannot match any legal SQL
+	 * identifier, so we don't bother to test for that case here.
 	 *
-	 * Should the user try to fool us by altering pg_attribute.attname for a
-	 * dropped column, we'll still catch it by virtue of the checks in
+	 * Should this somehow go wrong and we try to access a dropped column,
+	 * we'll still catch it by virtue of the checks in
 	 * get_rte_attribute_type(), which is called by make_var().  That
 	 * routine has to do a cache lookup anyway, so the check there is
 	 * cheap.
@@ -468,16 +504,14 @@ scanRTEForColumn(ParseState *pstate, RangeTblEntry *rte, char *colname)
 		attnum++;
 		if (strcmp(strVal(lfirst(c)), colname) == 0)
 		{
-			if (colname[0] == '.' &&	/* see note above */
-				get_rte_attribute_is_dropped(rte, attnum))
-				continue;
 			if (result)
 				ereport(ERROR,
 						(errcode(ERRCODE_AMBIGUOUS_COLUMN),
 						 errmsg("column reference \"%s\" is ambiguous",
 								colname)));
 			result = (Node *) make_var(pstate, rte, attnum);
-			rte->checkForRead = true;
+			/* Require read access */
+			rte->requiredPerms |= ACL_SELECT;
 		}
 	}
 
@@ -504,7 +538,8 @@ scanRTEForColumn(ParseState *pstate, RangeTblEntry *rte, char *colname)
 									 0, 0))
 			{
 				result = (Node *) make_var(pstate, rte, attnum);
-				rte->checkForRead = true;
+				/* Require read access */
+				rte->requiredPerms |= ACL_SELECT;
 			}
 		}
 	}
@@ -528,7 +563,7 @@ colNameToVar(ParseState *pstate, char *colname, bool localonly)
 
 	while (pstate != NULL)
 	{
-		List	   *ns;
+		ListCell   *ns;
 
 		/*
 		 * We need to look only at top-level namespace items, and even for
@@ -617,6 +652,81 @@ qualifiedNameToVar(ParseState *pstate,
 	return scanRTEForColumn(pstate, rte, colname);
 }
 
+/*
+ * buildRelationAliases
+ *		Construct the eref column name list for a relation RTE.
+ *		This code is also used for the case of a function RTE returning
+ *		a named composite type.
+ *
+ * tupdesc: the physical column information
+ * alias: the user-supplied alias, or NULL if none
+ * eref: the eref Alias to store column names in
+ *
+ * eref->colnames is filled in.  Also, alias->colnames is rebuilt to insert
+ * empty strings for any dropped columns, so that it will be one-to-one with
+ * physical column numbers.
+ */
+static void
+buildRelationAliases(TupleDesc tupdesc, Alias *alias, Alias *eref)
+{
+	int			maxattrs = tupdesc->natts;
+	ListCell   *aliaslc;
+	int			numaliases;
+	int			varattno;
+	int			numdropped = 0;
+
+	Assert(eref->colnames == NIL);
+
+	if (alias)
+	{
+		aliaslc = list_head(alias->colnames);
+		numaliases = list_length(alias->colnames);
+		/* We'll rebuild the alias colname list */
+		alias->colnames = NIL;
+	}
+	else
+	{
+		aliaslc = NULL;
+		numaliases = 0;
+	}
+
+	for (varattno = 0; varattno < maxattrs; varattno++)
+	{
+		Form_pg_attribute attr = tupdesc->attrs[varattno];
+		Value	   *attrname;
+
+		if (attr->attisdropped)
+		{
+			/* Always insert an empty string for a dropped column */
+			attrname = makeString(pstrdup(""));
+			if (aliaslc)
+				alias->colnames = lappend(alias->colnames, attrname);
+			numdropped++;
+		}
+		else if (aliaslc)
+		{
+			/* Use the next user-supplied alias */
+			attrname = (Value *) lfirst(aliaslc);
+			aliaslc = lnext(aliaslc);
+			alias->colnames = lappend(alias->colnames, attrname);
+		}
+		else
+		{
+			attrname = makeString(pstrdup(NameStr(attr->attname)));
+			/* we're done with the alias if any */
+		}
+
+		eref->colnames = lappend(eref->colnames, attrname);
+	}
+
+	/* Too many user-supplied aliases? */
+	if (aliaslc)
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
+				 errmsg("table \"%s\" has %d columns available but %d columns specified",
+				   eref->aliasname, maxattrs - numdropped, numaliases)));
+}
+
 /*
  * Add an entry for a relation to the pstate's range table (p_rtable).
  *
@@ -637,10 +747,6 @@ addRangeTableEntry(ParseState *pstate,
 	char	   *refname = alias ? alias->aliasname : relation->relname;
 	LOCKMODE	lockmode;
 	Relation	rel;
-	Alias	   *eref;
-	int			maxattrs;
-	int			numaliases;
-	int			varattno;
 
 	rte->rtekind = RTE_RELATION;
 	rte->alias = alias;
@@ -655,29 +761,12 @@ addRangeTableEntry(ParseState *pstate,
 	rel = heap_openrv(relation, lockmode);
 	rte->relid = RelationGetRelid(rel);
 
-	eref = alias ? (Alias *) copyObject(alias) : makeAlias(refname, NIL);
-	numaliases = length(eref->colnames);
-
 	/*
-	 * Since the rel is open anyway, let's check that the number of column
-	 * aliases is reasonable. - Thomas 2000-02-04
+	 * Build the list of effective column names using user-supplied
+	 * aliases and/or actual column names.
 	 */
-	maxattrs = RelationGetNumberOfAttributes(rel);
-	if (maxattrs < numaliases)
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
-				 errmsg("table \"%s\" has %d columns available but %d columns specified",
-				   RelationGetRelationName(rel), maxattrs, numaliases)));
-
-	/* fill in any unspecified alias columns using actual column names */
-	for (varattno = numaliases; varattno < maxattrs; varattno++)
-	{
-		char	   *attrname;
-
-		attrname = pstrdup(NameStr(rel->rd_att->attrs[varattno]->attname));
-		eref->colnames = lappend(eref->colnames, makeString(attrname));
-	}
-	rte->eref = eref;
+	rte->eref = makeAlias(refname, NIL);
+	buildRelationAliases(rel->rd_att, alias, rte->eref);
 
 	/*
 	 * Drop the rel refcount, but keep the access lock till end of
@@ -690,7 +779,7 @@ addRangeTableEntry(ParseState *pstate,
 	 * Flags:
 	 * - this RTE should be expanded to include descendant tables,
 	 * - this RTE is in the FROM clause,
-	 * - this RTE should be checked for read/write access rights.
+	 * - this RTE should be checked for appropriate access rights.
 	 *
 	 * The initial default on access checks is always check-for-READ-access,
 	 * which is the right thing for all except target tables.
@@ -698,10 +787,9 @@ addRangeTableEntry(ParseState *pstate,
 	 */
 	rte->inh = inh;
 	rte->inFromCl = inFromCl;
-	rte->checkForRead = true;
-	rte->checkForWrite = false;
 
-	rte->checkAsUser = InvalidOid;		/* not set-uid by default, either */
+	rte->requiredPerms = ACL_SELECT;
+	rte->checkAsUser = 0;		/* not set-uid by default, either */
 
 	/*
 	 * Add completed RTE to pstate's range table list, but not to join
@@ -732,10 +820,6 @@ addRangeTableEntryForRelation(ParseState *pstate,
 	char	   *refname = alias->aliasname;
 	LOCKMODE	lockmode;
 	Relation	rel;
-	Alias	   *eref;
-	int			maxattrs;
-	int			numaliases;
-	int			varattno;
 
 	rte->rtekind = RTE_RELATION;
 	rte->alias = alias;
@@ -750,29 +834,12 @@ addRangeTableEntryForRelation(ParseState *pstate,
 	rel = heap_open(relid, lockmode);
 	rte->relid = relid;
 
-	eref = (Alias *) copyObject(alias);
-	numaliases = length(eref->colnames);
-
 	/*
-	 * Since the rel is open anyway, let's check that the number of column
-	 * aliases is reasonable. - Thomas 2000-02-04
+	 * Build the list of effective column names using user-supplied
+	 * aliases and/or actual column names.
 	 */
-	maxattrs = RelationGetNumberOfAttributes(rel);
-	if (maxattrs < numaliases)
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
-				 errmsg("table \"%s\" has %d columns available but %d columns specified",
-				   RelationGetRelationName(rel), maxattrs, numaliases)));
-
-	/* fill in any unspecified alias columns using actual column names */
-	for (varattno = numaliases; varattno < maxattrs; varattno++)
-	{
-		char	   *attrname;
-
-		attrname = pstrdup(NameStr(rel->rd_att->attrs[varattno]->attname));
-		eref->colnames = lappend(eref->colnames, makeString(attrname));
-	}
-	rte->eref = eref;
+	rte->eref = makeAlias(refname, NIL);
+	buildRelationAliases(rel->rd_att, alias, rte->eref);
 
 	/*
 	 * Drop the rel refcount, but keep the access lock till end of
@@ -785,7 +852,7 @@ addRangeTableEntryForRelation(ParseState *pstate,
 	 * Flags:
 	 * - this RTE should be expanded to include descendant tables,
 	 * - this RTE is in the FROM clause,
-	 * - this RTE should be checked for read/write access rights.
+	 * - this RTE should be checked for appropriate access rights.
 	 *
 	 * The initial default on access checks is always check-for-READ-access,
 	 * which is the right thing for all except target tables.
@@ -793,10 +860,9 @@ addRangeTableEntryForRelation(ParseState *pstate,
 	 */
 	rte->inh = inh;
 	rte->inFromCl = inFromCl;
-	rte->checkForRead = true;
-	rte->checkForWrite = false;
 
-	rte->checkAsUser = InvalidOid;		/* not set-uid by default, either */
+	rte->requiredPerms = ACL_SELECT;
+	rte->checkAsUser = 0;		/* not set-uid by default, either */
 
 	/*
 	 * Add completed RTE to pstate's range table list, but not to join
@@ -825,7 +891,7 @@ addRangeTableEntryForSubquery(ParseState *pstate,
 	Alias	   *eref;
 	int			numaliases;
 	int			varattno;
-	List	   *tlistitem;
+	ListCell   *tlistitem;
 
 	rte->rtekind = RTE_SUBQUERY;
 	rte->relid = InvalidOid;
@@ -833,7 +899,7 @@ addRangeTableEntryForSubquery(ParseState *pstate,
 	rte->alias = alias;
 
 	eref = copyObject(alias);
-	numaliases = length(eref->colnames);
+	numaliases = list_length(eref->colnames);
 
 	/* fill in any unspecified alias columns */
 	varattno = 0;
@@ -865,17 +931,16 @@ addRangeTableEntryForSubquery(ParseState *pstate,
 	 * Flags:
 	 * - this RTE should be expanded to include descendant tables,
 	 * - this RTE is in the FROM clause,
-	 * - this RTE should be checked for read/write access rights.
+	 * - this RTE should be checked for appropriate access rights.
 	 *
 	 * Subqueries are never checked for access rights.
 	 *----------
 	 */
 	rte->inh = false;			/* never true for subqueries */
 	rte->inFromCl = inFromCl;
-	rte->checkForRead = false;
-	rte->checkForWrite = false;
 
-	rte->checkAsUser = InvalidOid;
+	rte->requiredPerms = 0;
+	rte->checkAsUser = 0;
 
 	/*
 	 * Add completed RTE to pstate's range table list, but not to join
@@ -901,12 +966,10 @@ addRangeTableEntryForFunction(ParseState *pstate,
 {
 	RangeTblEntry *rte = makeNode(RangeTblEntry);
 	Oid			funcrettype = exprType(funcexpr);
-	char		functyptype;
+	TypeFuncClass functypclass;
 	Alias	   *alias = rangefunc->alias;
 	List	   *coldeflist = rangefunc->coldeflist;
 	Alias	   *eref;
-	int			numaliases;
-	int			varattno;
 
 	rte->rtekind = RTE_FUNCTION;
 	rte->relid = InvalidOid;
@@ -915,11 +978,9 @@ addRangeTableEntryForFunction(ParseState *pstate,
 	rte->coldeflist = coldeflist;
 	rte->alias = alias;
 
-	eref = alias ? (Alias *) copyObject(alias) : makeAlias(funcname, NIL);
+	eref = makeAlias(alias ? alias->aliasname : funcname, NIL);
 	rte->eref = eref;
 
-	numaliases = length(eref->colnames);
-
 	/*
 	 * Now determine if the function returns a simple or composite type,
 	 * and check/add column aliases.
@@ -947,19 +1008,15 @@ addRangeTableEntryForFunction(ParseState *pstate,
 					 errmsg("a column definition list is required for functions returning \"record\"")));
 	}
 
-	functyptype = get_typtype(funcrettype);
+	functypclass = get_type_func_class(funcrettype);
 
-	if (functyptype == 'c')
+	if (functypclass == TYPEFUNC_COMPOSITE)
 	{
-		/*
-		 * Named composite data type, i.e. a table's row type
-		 */
+		/* Composite data type, e.g. a table's row type */
 		Oid			funcrelid = typeidTypeRelid(funcrettype);
 		Relation	rel;
-		int			maxattrs;
 
-		if (!OidIsValid(funcrelid))		/* shouldn't happen if typtype is
-										 * 'c' */
+		if (!OidIsValid(funcrelid))		/* shouldn't happen */
 			elog(ERROR, "invalid typrelid for complex type %u", funcrettype);
 
 		/*
@@ -968,26 +1025,8 @@ addRangeTableEntryForFunction(ParseState *pstate,
 		 */
 		rel = relation_open(funcrelid, AccessShareLock);
 
-		/*
-		 * Since the rel is open anyway, let's check that the number of
-		 * column aliases is reasonable.
-		 */
-		maxattrs = RelationGetNumberOfAttributes(rel);
-		if (maxattrs < numaliases)
-			ereport(ERROR,
-					(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
-					 errmsg("table \"%s\" has %d columns available but %d columns specified",
-							RelationGetRelationName(rel),
-							maxattrs, numaliases)));
-
-		/* fill in alias columns using actual column names */
-		for (varattno = numaliases; varattno < maxattrs; varattno++)
-		{
-			char	   *attrname;
-
-			attrname = pstrdup(NameStr(rel->rd_att->attrs[varattno]->attname));
-			eref->colnames = lappend(eref->colnames, makeString(attrname));
-		}
+		/* Build the column alias list */
+		buildRelationAliases(rel->rd_att, alias, eref);
 
 		/*
 		 * Drop the rel refcount, but keep the access lock till end of
@@ -996,26 +1035,27 @@ addRangeTableEntryForFunction(ParseState *pstate,
 		 */
 		relation_close(rel, NoLock);
 	}
-	else if (functyptype == 'b' || functyptype == 'd')
+	else if (functypclass == TYPEFUNC_SCALAR)
 	{
-		/*
-		 * Must be a base data type, i.e. scalar. Just add one alias
-		 * column named for the function.
-		 */
-		if (numaliases > 1)
-			ereport(ERROR,
-					(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
-			  errmsg("too many column aliases specified for function %s",
-					 funcname)));
-		if (numaliases == 0)
-			eref->colnames = makeList1(makeString(eref->aliasname));
+		/* Base data type, i.e. scalar */
+		/* Just add one alias column named for the function. */
+		if (alias && alias->colnames != NIL)
+		{
+			if (list_length(alias->colnames) != 1)
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
+						 errmsg("too many column aliases specified for function %s",
+								funcname)));
+			eref->colnames = copyObject(alias->colnames);
+		}
+		else
+			eref->colnames = list_make1(makeString(eref->aliasname));
 	}
-	else if (functyptype == 'p' && funcrettype == RECORDOID)
+	else if (functypclass == TYPEFUNC_RECORD)
 	{
-		List	   *col;
+		ListCell   *col;
 
 		/* Use the column definition list to form the alias list */
-		eref->colnames = NIL;
 		foreach(col, coldeflist)
 		{
 			ColumnDef  *n = lfirst(col);
@@ -1028,22 +1068,24 @@ addRangeTableEntryForFunction(ParseState *pstate,
 	else
 		ereport(ERROR,
 				(errcode(ERRCODE_DATATYPE_MISMATCH),
-			errmsg("function \"%s\" in FROM has unsupported return type",
-				   funcname)));
+			errmsg("function \"%s\" in FROM has unsupported return type %s",
+				   funcname, format_type_be(funcrettype))));
 
 	/*----------
 	 * Flags:
 	 * - this RTE should be expanded to include descendant tables,
 	 * - this RTE is in the FROM clause,
-	 * - this RTE should be checked for read/write access rights.
+	 * - this RTE should be checked for appropriate access rights.
+	 *
+	 * Functions are never checked for access rights (at least, not by
+	 * the RTE permissions mechanism).
 	 *----------
 	 */
 	rte->inh = false;			/* never true for functions */
 	rte->inFromCl = inFromCl;
-	rte->checkForRead = true;
-	rte->checkForWrite = false;
 
-	rte->checkAsUser = InvalidOid;
+	rte->requiredPerms = 0;
+	rte->checkAsUser = 0;
 
 	/*
 	 * Add completed RTE to pstate's range table list, but not to join
@@ -1080,15 +1122,12 @@ addRangeTableEntryForJoin(ParseState *pstate,
 	rte->alias = alias;
 
 	eref = alias ? (Alias *) copyObject(alias) : makeAlias("unnamed_join", NIL);
-	numaliases = length(eref->colnames);
+	numaliases = list_length(eref->colnames);
 
 	/* fill in any unspecified alias columns */
-	if (numaliases < length(colnames))
-	{
-		while (numaliases-- > 0)
-			colnames = lnext(colnames);
-		eref->colnames = nconc(eref->colnames, colnames);
-	}
+	if (numaliases < list_length(colnames))
+		eref->colnames = list_concat(eref->colnames,
+								   list_copy_tail(colnames, numaliases));
 
 	rte->eref = eref;
 
@@ -1096,17 +1135,16 @@ addRangeTableEntryForJoin(ParseState *pstate,
 	 * Flags:
 	 * - this RTE should be expanded to include descendant tables,
 	 * - this RTE is in the FROM clause,
-	 * - this RTE should be checked for read/write access rights.
+	 * - this RTE should be checked for appropriate access rights.
 	 *
 	 * Joins are never checked for access rights.
 	 *----------
 	 */
 	rte->inh = false;			/* never true for joins */
 	rte->inFromCl = inFromCl;
-	rte->checkForRead = false;
-	rte->checkForWrite = false;
 
-	rte->checkAsUser = InvalidOid;
+	rte->requiredPerms = 0;
+	rte->checkAsUser = 0;
 
 	/*
 	 * Add completed RTE to pstate's range table list, but not to join
@@ -1129,7 +1167,7 @@ isForUpdate(ParseState *pstate, char *refname)
 	{
 		if (pstate->p_forUpdate != NIL)
 		{
-			if (lfirst(pstate->p_forUpdate) == NULL)
+			if (linitial(pstate->p_forUpdate) == NULL)
 			{
 				/* all tables used in query */
 				return true;
@@ -1137,7 +1175,7 @@ isForUpdate(ParseState *pstate, char *refname)
 			else
 			{
 				/* just the named tables */
-				List	   *l;
+				ListCell   *l;
 
 				foreach(l, pstate->p_forUpdate)
 				{
@@ -1191,83 +1229,48 @@ addImplicitRTE(ParseState *pstate, RangeVar *relation)
 	return rte;
 }
 
-/* expandRTE()
+/*
+ * expandRTE -- expand the columns of a rangetable entry
  *
- * Given a rangetable entry, create lists of its column names (aliases if
- * provided, else real names) and Vars for each column.  Only user columns
- * are considered, since this is primarily used to expand '*' and determine
- * the contents of JOIN tables.
+ * This creates lists of an RTE's column names (aliases if provided, else
+ * real names) and Vars for each column.  Only user columns are considered.
+ * If include_dropped is FALSE then dropped columns are omitted from the
+ * results.  If include_dropped is TRUE then empty strings and NULL constants
+ * (not Vars!) are returned for dropped columns.
  *
+ * The target RTE is the rtindex'th entry of rtable.  (The whole rangetable
+ * must be passed since we need it to determine dropped-ness for JOIN columns.)
+ * sublevels_up is the varlevelsup value to use in the created Vars.
+ *
+ * The output lists go into *colnames and *colvars.
  * If only one of the two kinds of output list is needed, pass NULL for the
  * output pointer for the unwanted one.
  */
 void
-expandRTE(ParseState *pstate, RangeTblEntry *rte,
+expandRTE(List *rtable, int rtindex, int sublevels_up,
+		  bool include_dropped,
 		  List **colnames, List **colvars)
 {
-	int			rtindex,
-				sublevels_up,
-				varattno;
+	RangeTblEntry *rte = rt_fetch(rtindex, rtable);
+	int			varattno;
 
 	if (colnames)
 		*colnames = NIL;
 	if (colvars)
 		*colvars = NIL;
 
-	/* Need the RT index of the entry for creating Vars */
-	rtindex = RTERangeTablePosn(pstate, rte, &sublevels_up);
-
 	switch (rte->rtekind)
 	{
 		case RTE_RELATION:
-			{
-				/* Ordinary relation RTE */
-				Relation	rel;
-				int			maxattrs;
-				int			numaliases;
-
-				rel = heap_open(rte->relid, AccessShareLock);
-				maxattrs = RelationGetNumberOfAttributes(rel);
-				numaliases = length(rte->eref->colnames);
-
-				for (varattno = 0; varattno < maxattrs; varattno++)
-				{
-					Form_pg_attribute attr = rel->rd_att->attrs[varattno];
-
-					if (attr->attisdropped)
-						continue;
-
-					if (colnames)
-					{
-						char	   *label;
-
-						if (varattno < numaliases)
-							label = strVal(nth(varattno, rte->eref->colnames));
-						else
-							label = NameStr(attr->attname);
-						*colnames = lappend(*colnames, makeString(pstrdup(label)));
-					}
-
-					if (colvars)
-					{
-						Var		   *varnode;
-
-						varnode = makeVar(rtindex, attr->attnum,
-										  attr->atttypid, attr->atttypmod,
-										  sublevels_up);
-
-						*colvars = lappend(*colvars, varnode);
-					}
-				}
-
-				heap_close(rel, AccessShareLock);
-			}
+			/* Ordinary relation RTE */
+			expandRelation(rte->relid, rte->eref, rtindex, sublevels_up,
+						   include_dropped, colnames, colvars);
 			break;
 		case RTE_SUBQUERY:
 			{
 				/* Subquery RTE */
-				List	   *aliasp = rte->eref->colnames;
-				List	   *tlistitem;
+				ListCell   *aliasp_item = list_head(rte->eref->colnames);
+				ListCell   *tlistitem;
 
 				varattno = 0;
 				foreach(tlistitem, rte->subquery->targetList)
@@ -1282,10 +1285,10 @@ expandRTE(ParseState *pstate, RangeTblEntry *rte,
 					if (colnames)
 					{
 						/* Assume there is one alias per target item */
-						char	   *label = strVal(lfirst(aliasp));
+						char	   *label = strVal(lfirst(aliasp_item));
 
 						*colnames = lappend(*colnames, makeString(pstrdup(label)));
-						aliasp = lnext(aliasp);
+						aliasp_item = lnext(aliasp_item);
 					}
 
 					if (colvars)
@@ -1306,70 +1309,30 @@ expandRTE(ParseState *pstate, RangeTblEntry *rte,
 			{
 				/* Function RTE */
 				Oid			funcrettype = exprType(rte->funcexpr);
-				char		functyptype = get_typtype(funcrettype);
-				List	   *coldeflist = rte->coldeflist;
+				TypeFuncClass functypclass = get_type_func_class(funcrettype);
 
-				if (functyptype == 'c')
+				if (functypclass == TYPEFUNC_COMPOSITE)
 				{
 					/*
-					 * Composite data type, i.e. a table's row type Same
-					 * as ordinary relation RTE
+					 * Composite data type, i.e. a table's row type
+					 *
+					 * Same as ordinary relation RTE
 					 */
 					Oid			funcrelid = typeidTypeRelid(funcrettype);
-					Relation	rel;
-					int			maxattrs;
-					int			numaliases;
 
 					if (!OidIsValid(funcrelid)) /* shouldn't happen */
 						elog(ERROR, "invalid typrelid for complex type %u",
 							 funcrettype);
 
-					rel = relation_open(funcrelid, AccessShareLock);
-					maxattrs = RelationGetNumberOfAttributes(rel);
-					numaliases = length(rte->eref->colnames);
-
-					for (varattno = 0; varattno < maxattrs; varattno++)
-					{
-						Form_pg_attribute attr = rel->rd_att->attrs[varattno];
-
-						if (attr->attisdropped)
-							continue;
-
-						if (colnames)
-						{
-							char	   *label;
-
-							if (varattno < numaliases)
-								label = strVal(nth(varattno, rte->eref->colnames));
-							else
-								label = NameStr(attr->attname);
-							*colnames = lappend(*colnames, makeString(pstrdup(label)));
-						}
-
-						if (colvars)
-						{
-							Var		   *varnode;
-
-							varnode = makeVar(rtindex,
-											  attr->attnum,
-											  attr->atttypid,
-											  attr->atttypmod,
-											  sublevels_up);
-
-							*colvars = lappend(*colvars, varnode);
-						}
-					}
-
-					relation_close(rel, AccessShareLock);
+					expandRelation(funcrelid, rte->eref, rtindex, sublevels_up,
+								   include_dropped, colnames, colvars);
 				}
-				else if (functyptype == 'b' || functyptype == 'd')
+				else if (functypclass == TYPEFUNC_SCALAR)
 				{
-					/*
-					 * Must be a base data type, i.e. scalar
-					 */
+					/* Base data type, i.e. scalar */
 					if (colnames)
 						*colnames = lappend(*colnames,
-											lfirst(rte->eref->colnames));
+										  linitial(rte->eref->colnames));
 
 					if (colvars)
 					{
@@ -1382,9 +1345,10 @@ expandRTE(ParseState *pstate, RangeTblEntry *rte,
 						*colvars = lappend(*colvars, varnode);
 					}
 				}
-				else if (functyptype == 'p' && funcrettype == RECORDOID)
+				else if (functypclass == TYPEFUNC_RECORD)
 				{
-					List	   *col;
+					List	   *coldeflist = rte->coldeflist;
+					ListCell   *col;
 					int			attnum = 0;
 
 					foreach(col, coldeflist)
@@ -1418,47 +1382,75 @@ expandRTE(ParseState *pstate, RangeTblEntry *rte,
 					}
 				}
 				else
-					ereport(ERROR,
-							(errcode(ERRCODE_DATATYPE_MISMATCH),
-							 errmsg("function in FROM has unsupported return type")));
+				{
+					/* addRangeTableEntryForFunction should've caught this */
+					elog(ERROR, "function in FROM has unsupported return type");
+				}
 			}
 			break;
 		case RTE_JOIN:
 			{
 				/* Join RTE */
-				List	   *aliasp = rte->eref->colnames;
-				List	   *aliasvars = rte->joinaliasvars;
+				ListCell   *colname;
+				ListCell   *aliasvar;
+
+				Assert(list_length(rte->eref->colnames) == list_length(rte->joinaliasvars));
 
 				varattno = 0;
-				while (aliasp)
+				forboth(colname, rte->eref->colnames, aliasvar, rte->joinaliasvars)
 				{
-					Assert(aliasvars);
 					varattno++;
 
+					/*
+					 * During ordinary parsing, there will never be any
+					 * deleted columns in the join; but we have to check
+					 * since this routine is also used by the rewriter,
+					 * and joins found in stored rules might have join
+					 * columns for since-deleted columns.
+					 */
+					if (get_rte_attribute_is_dropped(rtable, rtindex,
+													 varattno))
+					{
+						if (include_dropped)
+						{
+							if (colnames)
+								*colnames = lappend(*colnames,
+												makeString(pstrdup("")));
+							if (colvars)
+							{
+								/*
+								 * can't use atttypid here, but it doesn't
+								 * really matter what type the Const
+								 * claims to be.
+								 */
+								*colvars = lappend(*colvars,
+												 makeNullConst(INT4OID));
+							}
+						}
+						continue;
+					}
+
 					if (colnames)
 					{
-						char	   *label = strVal(lfirst(aliasp));
+						char	   *label = strVal(lfirst(colname));
 
-						*colnames = lappend(*colnames, makeString(pstrdup(label)));
+						*colnames = lappend(*colnames,
+											makeString(pstrdup(label)));
 					}
 
 					if (colvars)
 					{
-						Node	   *aliasvar = (Node *) lfirst(aliasvars);
+						Node	   *avar = (Node *) lfirst(aliasvar);
 						Var		   *varnode;
 
 						varnode = makeVar(rtindex, varattno,
-										  exprType(aliasvar),
-										  exprTypmod(aliasvar),
+										  exprType(avar),
+										  exprTypmod(avar),
 										  sublevels_up);
 
 						*colvars = lappend(*colvars, varnode);
 					}
-
-					aliasp = lnext(aliasp);
-					aliasvars = lnext(aliasvars);
 				}
-				Assert(aliasvars == NIL);
 			}
 			break;
 		default:
@@ -1466,24 +1458,91 @@ expandRTE(ParseState *pstate, RangeTblEntry *rte,
 	}
 }
 
+/*
+ * expandRelation -- expandRTE subroutine
+ */
+static void
+expandRelation(Oid relid, Alias *eref, int rtindex, int sublevels_up,
+			   bool include_dropped,
+			   List **colnames, List **colvars)
+{
+	Relation	rel;
+	int			varattno;
+	int			maxattrs;
+	int			numaliases;
+
+	rel = relation_open(relid, AccessShareLock);
+	maxattrs = RelationGetNumberOfAttributes(rel);
+	numaliases = list_length(eref->colnames);
+
+	for (varattno = 0; varattno < maxattrs; varattno++)
+	{
+		Form_pg_attribute attr = rel->rd_att->attrs[varattno];
+
+		if (attr->attisdropped)
+		{
+			if (include_dropped)
+			{
+				if (colnames)
+					*colnames = lappend(*colnames, makeString(pstrdup("")));
+				if (colvars)
+				{
+					/*
+					 * can't use atttypid here, but it doesn't really
+					 * matter what type the Const claims to be.
+					 */
+					*colvars = lappend(*colvars, makeNullConst(INT4OID));
+				}
+			}
+			continue;
+		}
+
+		if (colnames)
+		{
+			char	   *label;
+
+			if (varattno < numaliases)
+				label = strVal(list_nth(eref->colnames, varattno));
+			else
+				label = NameStr(attr->attname);
+			*colnames = lappend(*colnames, makeString(pstrdup(label)));
+		}
+
+		if (colvars)
+		{
+			Var		   *varnode;
+
+			varnode = makeVar(rtindex, attr->attnum,
+							  attr->atttypid, attr->atttypmod,
+							  sublevels_up);
+
+			*colvars = lappend(*colvars, varnode);
+		}
+	}
+
+	relation_close(rel, AccessShareLock);
+}
+
 /*
  * expandRelAttrs -
  *	  Workhorse for "*" expansion: produce a list of targetentries
  *	  for the attributes of the rte
  */
 List *
-expandRelAttrs(ParseState *pstate, RangeTblEntry *rte)
+expandRelAttrs(ParseState *pstate, List *rtable, int rtindex, int sublevels_up)
 {
 	List	   *names,
 			   *vars;
+	ListCell   *name,
+			   *var;
 	List	   *te_list = NIL;
 
-	expandRTE(pstate, rte, &names, &vars);
+	expandRTE(rtable, rtindex, sublevels_up, false, &names, &vars);
 
-	while (names)
+	forboth(name, names, var, vars)
 	{
-		char	   *label = strVal(lfirst(names));
-		Node	   *varnode = (Node *) lfirst(vars);
+		char	   *label = strVal(lfirst(name));
+		Node	   *varnode = (Node *) lfirst(var);
 		TargetEntry *te = makeNode(TargetEntry);
 
 		te->resdom = makeResdom((AttrNumber) pstate->p_next_resno++,
@@ -1493,12 +1552,10 @@ expandRelAttrs(ParseState *pstate, RangeTblEntry *rte)
 								false);
 		te->expr = (Expr *) varnode;
 		te_list = lappend(te_list, te);
-
-		names = lnext(names);
-		vars = lnext(vars);
 	}
 
-	Assert(vars == NIL);		/* lists not same length? */
+	Assert(name == NULL && var == NULL);		/* lists not the same
+												 * length? */
 
 	return te_list;
 }
@@ -1524,8 +1581,8 @@ get_rte_attribute_name(RangeTblEntry *rte, AttrNumber attnum)
 	 * If there is a user-written column alias, use it.
 	 */
 	if (rte->alias &&
-		attnum > 0 && attnum <= length(rte->alias->colnames))
-		return strVal(nth(attnum - 1, rte->alias->colnames));
+		attnum > 0 && attnum <= list_length(rte->alias->colnames))
+		return strVal(list_nth(rte->alias->colnames, attnum - 1));
 
 	/*
 	 * If the RTE is a relation, go to the system catalogs not the
@@ -1540,8 +1597,8 @@ get_rte_attribute_name(RangeTblEntry *rte, AttrNumber attnum)
 	 * Otherwise use the column name from eref.  There should always be
 	 * one.
 	 */
-	if (attnum > 0 && attnum <= length(rte->eref->colnames))
-		return strVal(nth(attnum - 1, rte->eref->colnames));
+	if (attnum > 0 && attnum <= list_length(rte->eref->colnames))
+		return strVal(list_nth(rte->eref->colnames, attnum - 1));
 
 	/* else caller gave us a bogus attnum */
 	elog(ERROR, "invalid attnum %d for rangetable entry %s",
@@ -1606,14 +1663,15 @@ get_rte_attribute_type(RangeTblEntry *rte, AttrNumber attnum,
 			{
 				/* Function RTE */
 				Oid			funcrettype = exprType(rte->funcexpr);
-				char		functyptype = get_typtype(funcrettype);
+				TypeFuncClass functypclass = get_type_func_class(funcrettype);
 				List	   *coldeflist = rte->coldeflist;
 
-				if (functyptype == 'c')
+				if (functypclass == TYPEFUNC_COMPOSITE)
 				{
 					/*
-					 * Composite data type, i.e. a table's row type Same
-					 * as ordinary relation RTE
+					 * Composite data type, i.e. a table's row type
+					 *
+					 * Same as ordinary relation RTE
 					 */
 					Oid			funcrelid = typeidTypeRelid(funcrettype);
 					HeapTuple	tp;
@@ -1646,25 +1704,24 @@ get_rte_attribute_type(RangeTblEntry *rte, AttrNumber attnum,
 					*vartypmod = att_tup->atttypmod;
 					ReleaseSysCache(tp);
 				}
-				else if (functyptype == 'b' || functyptype == 'd')
+				else if (functypclass == TYPEFUNC_SCALAR)
 				{
-					/*
-					 * Must be a base data type, i.e. scalar
-					 */
+					/* Base data type, i.e. scalar */
 					*vartype = funcrettype;
 					*vartypmod = -1;
 				}
-				else if (functyptype == 'p' && funcrettype == RECORDOID)
+				else if (functypclass == TYPEFUNC_RECORD)
 				{
-					ColumnDef  *colDef = nth(attnum - 1, coldeflist);
+					ColumnDef  *colDef = list_nth(coldeflist, attnum - 1);
 
 					*vartype = typenameTypeId(colDef->typename);
 					*vartypmod = colDef->typename->typmod;
 				}
 				else
-					ereport(ERROR,
-							(errcode(ERRCODE_DATATYPE_MISMATCH),
-							 errmsg("function in FROM has unsupported return type")));
+				{
+					/* addRangeTableEntryForFunction should've caught this */
+					elog(ERROR, "function in FROM has unsupported return type");
+				}
 			}
 			break;
 		case RTE_JOIN:
@@ -1675,8 +1732,8 @@ get_rte_attribute_type(RangeTblEntry *rte, AttrNumber attnum,
 				 */
 				Node	   *aliasvar;
 
-				Assert(attnum > 0 && attnum <= length(rte->joinaliasvars));
-				aliasvar = (Node *) nth(attnum - 1, rte->joinaliasvars);
+				Assert(attnum > 0 && attnum <= list_length(rte->joinaliasvars));
+				aliasvar = (Node *) list_nth(rte->joinaliasvars, attnum - 1);
 				*vartype = exprType(aliasvar);
 				*vartypmod = exprTypmod(aliasvar);
 			}
@@ -1690,16 +1747,20 @@ get_rte_attribute_type(RangeTblEntry *rte, AttrNumber attnum,
  * get_rte_attribute_is_dropped
  *		Check whether attempted attribute ref is to a dropped column
  */
-static bool
-get_rte_attribute_is_dropped(RangeTblEntry *rte, AttrNumber attnum)
+bool
+get_rte_attribute_is_dropped(List *rtable, int rtindex, AttrNumber attnum)
 {
+	RangeTblEntry *rte = rt_fetch(rtindex, rtable);
 	bool		result;
 
 	switch (rte->rtekind)
 	{
 		case RTE_RELATION:
 			{
-				/* Plain relation RTE --- get the attribute's type info */
+				/*
+				 * Plain relation RTE --- get the attribute's catalog
+				 * entry
+				 */
 				HeapTuple	tp;
 				Form_pg_attribute att_tup;
 
@@ -1716,10 +1777,40 @@ get_rte_attribute_is_dropped(RangeTblEntry *rte, AttrNumber attnum)
 			}
 			break;
 		case RTE_SUBQUERY:
-		case RTE_JOIN:
-			/* Subselect and join RTEs never have dropped columns */
+			/* Subselect RTEs never have dropped columns */
 			result = false;
 			break;
+		case RTE_JOIN:
+			{
+				/*
+				 * A join RTE would not have dropped columns when
+				 * constructed, but one in a stored rule might contain
+				 * columns that were dropped from the underlying tables,
+				 * if said columns are nowhere explicitly referenced in
+				 * the rule.  So we have to recursively look at the
+				 * referenced column.
+				 */
+				Var		   *aliasvar;
+
+				if (attnum <= 0 ||
+					attnum > list_length(rte->joinaliasvars))
+					elog(ERROR, "invalid varattno %d", attnum);
+				aliasvar = (Var *) list_nth(rte->joinaliasvars, attnum - 1);
+
+				/*
+				 * If the list item isn't a simple Var, then it must
+				 * represent a merged column, ie a USING column, and so it
+				 * couldn't possibly be dropped (since it's referenced in
+				 * the join clause).
+				 */
+				if (!IsA(aliasvar, Var))
+					result = false;
+				else
+					result = get_rte_attribute_is_dropped(rtable,
+														  aliasvar->varno,
+													 aliasvar->varattno);
+			}
+			break;
 		case RTE_FUNCTION:
 			{
 				/* Function RTE */
@@ -1729,8 +1820,9 @@ get_rte_attribute_is_dropped(RangeTblEntry *rte, AttrNumber attnum)
 				if (OidIsValid(funcrelid))
 				{
 					/*
-					 * Composite data type, i.e. a table's row type Same
-					 * as ordinary relation RTE
+					 * Composite data type, i.e. a table's row type
+					 *
+					 * Same as ordinary relation RTE
 					 */
 					HeapTuple	tp;
 					Form_pg_attribute att_tup;
@@ -1768,17 +1860,17 @@ get_rte_attribute_is_dropped(RangeTblEntry *rte, AttrNumber attnum)
  *
  * Returns NULL if resno is not present in list.
  *
- * Note: we need to search, rather than just indexing with nth(), because
- * not all tlists are sorted by resno.
+ * Note: we need to search, rather than just indexing with list_nth(),
+ * because not all tlists are sorted by resno.
  */
 TargetEntry *
 get_tle_by_resno(List *tlist, AttrNumber resno)
 {
-	List	   *i;
+	ListCell   *l;
 
-	foreach(i, tlist)
+	foreach(l, tlist)
 	{
-		TargetEntry *tle = (TargetEntry *) lfirst(i);
+		TargetEntry *tle = (TargetEntry *) lfirst(l);
 
 		if (tle->resdom->resno == resno)
 			return tle;
@@ -1901,7 +1993,7 @@ static void
 warnAutoRange(ParseState *pstate, RangeVar *relation)
 {
 	bool		foundInFromCl = false;
-	List	   *temp;
+	ListCell   *temp;
 
 	if (!add_missing_from)
 	{
